"""
Copyright 2024 Eviden
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an 'AS IS' BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.


This module provides a number of utility functions that are used by the CLI
"""

import os
from typing import Optional
import paramiko
from hid_data_transfer_lib.exceptions.hid_dt_exceptions import (
    HidDataTransferException
)


def order_process_group(connections):
    """Order a sequence of processors from an array of connections"""
    sequence = []
    sequence.append(connections[0])
    for pair in connections[1:]:
        source_id = pair[0]
        destination_id = pair[1]
        source_index = get_index(source_id, sequence)
        destination_index = get_index(destination_id, sequence)
        if source_index is not None:
            if destination_index is None:
                sequence.insert(source_index + 1, destination_id)
            else:
                if source_index > destination_index:
                    sequence = swap(sequence, source_index, destination_index)
        else:
            if destination_index is None:
                sequence.append(pair)
            else:
                insert_index = 0 if destination_index == 0 else \
                    destination_index - 1
                sequence.insert(insert_index, source_id)
    return flatten(sequence)


def flatten(sequence):
    """Flatten a nested list into a single list"""
    result = []
    for elem in sequence:
        if isinstance(elem, list):
            for item in elem:
                result.append(item)
        else:
            result.append(elem)
    return result


def swap(sequence, index1, index2):
    """Swap two elements in a list at the given indices"""
    sequence[index1], sequence[index2] = sequence[index2], sequence[index1]
    return sequence


def get_index(target, sequence) -> Optional[int]:
    """Get the index of an element in a list or nested list"""
    for index, elem in enumerate(sequence):
        if isinstance(elem, list):
            if target in elem:
                return index
        else:
            if target == elem:
                return index
    return None


def sftp_put(
    server,
    source,
    target,
    username,
    private_key=None,
    target_permissions=None,
    target_ownership=None,
):
    """Transfer a file from the local machine to a remote server using SFTP"""
    if source.startswith("~"):
        source = os.path.expanduser(os.path.join(*source.split("/")))
    if private_key.startswith("~"):
        private_key = os.path.expanduser(os.path.join(*private_key.split("/")))
    with paramiko.SSHClient() as ssh:
        ssh.load_host_keys(os.path.expanduser(
            os.path.join("~", ".ssh", "known_hosts")))
        if not private_key:
            ssh.connect(server, username=username)
        else:
            ssh.connect(server, username=username, key_filename=private_key)
        with ssh.open_sftp() as sftp:
            sftp.put(source, target)
            if target_permissions:
                command = f"chmod {target_permissions} {target}"
                ssh.exec_command(command)
            if target_ownership:
                command = f"chown {target_ownership} {target}"
                ssh.exec_command(command)


def sftp_get(server, source, target, username, private_key=None):
    """Transfer a file from a remote server to the local machine using SFTP"""
    if target.startswith("~"):
        target = os.path.expanduser(os.path.join(*target.split("/")))
    if private_key.startswith("~"):
        private_key = os.path.expanduser(os.path.join(*private_key.split("/")))
    with paramiko.SSHClient() as ssh:
        ssh.load_host_keys(os.path.expanduser(
            os.path.join("~", ".ssh", "known_hosts")))
        if not private_key:
            ssh.connect(server, username=username)
        else:
            ssh.connect(server, username=username, key_filename=private_key)
        with ssh.open_sftp() as sftp:
            sftp.get(remotepath=source, localpath=target)


def ssh_delete_file(file, server, username, private_key=None):
    """Delete a file on a remote server using SSH"""
    command = f"rm {file}"
    __ssh_execute_command(server, username, command, private_key)


def ssh_create_folder(folder, server, username, private_key=None):
    """Delete a file on a remote server using SSH"""
    command = f"mkdir -p {folder}"
    __ssh_execute_command(server, username, command, private_key)

    # set ownership and permissions
    ownership_command = f"chown {username}:nifi {folder}"
    __ssh_execute_command(server, username, ownership_command, private_key)
    permissions_command = f"chmod 775 {folder}"
    __ssh_execute_command(server, username, permissions_command, private_key)


def ssh_delete_folder(folder, server, username, private_key=None):
    """Delete a file on a remote server using SSH"""
    command = f"rm -rf {folder}"
    __ssh_execute_command(server, username, command, private_key)


def __ssh_execute_command(server, username, command, private_key=None):
    """Execute a command on a remote server using SSH"""
    if private_key.startswith("~"):
        private_key = os.path.expanduser(os.path.join(*private_key.split("/")))
    with paramiko.SSHClient() as ssh:
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.load_host_keys(os.path.expanduser(
            os.path.join("~", ".ssh", "known_hosts")))
        if not private_key:
            ssh.connect(server, username=username)
        else:
            ssh.connect(server, username=username, key_filename=private_key)
        _, _, stderr = ssh.exec_command(command)
        error = stderr.read().decode("utf-8")
        if error:
            raise HidDataTransferException(
                f"Error '{error}' executing command '{command}'"
            )


# list all files in path starting with prefix
def ssh_list(path, server, username, private_key=None, prefix=None) -> list:
    """List files on a remote server using SSH"""
    if private_key.startswith("~"):
        private_key = os.path.expanduser(os.path.join(*private_key.split("/")))
    with paramiko.SSHClient() as ssh:
        ssh.load_host_keys(os.path.expanduser(
            os.path.join("~", ".ssh", "known_hosts")))
        if not private_key:
            ssh.connect(server, username=username)
        else:
            ssh.connect(server, username=username, key_filename=private_key)
        if prefix is not None:
            command = (
                f'find {path} -maxdepth 1 '
                f'-name "{prefix}*" -type f -printf "%p\n"'
            )
        else:
            command = f'find {path} -maxdepth 1 -type f -printf "%p\n"'
        _, stdout, _ = ssh.exec_command(command)

        # read lines from stdout in a list
        files = []
        for line in stdout:
            files.append(line.strip())
        return files
